import torch
import math


def gelu_new(x):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT). Also see
    the Gaussian Error Linear Units paper: https://arxiv.org/abs/1606.08415
    """
    return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))

def gelu_fast(x):
    return 0.5 * x * (1.0 + torch.tanh(x * 0.7978845608 * (1.0 + 0.044715 * x * x)))

def mish(x):
    return x * torch.tanh(torch.nn.functional.softplus(x))

ACT2FN = {
    "relu": torch.nn.functional.relu,
    "gelu": torch.nn.functional.gelu,
    "tanh": torch.tanh,
    "gelu_new": gelu_new,
    "gelu_fast": gelu_fast,
    "mish": mish,
    "sigmoid": torch.sigmoid,
}



import os, json, argparse, string, subprocess, re
from dataclasses import dataclass

from numba import jit
from numba.typed import Dict, List
from numba.types import unicode_type, DictType

import numpy as np
import torch
import requests

from . import tokenizers, model


class pretrained:
    """
    Initializes AbLang for heavy or light chains.    
    """
    
    def __init__(self, chain="heavy", model_folder="download", random_init=False, ncpu=7, device='cpu'):
        super().__init__()
        
        self.used_device = torch.device(device)
        
        if model_folder == "download":
            # Download model and save to specific place - if already downloaded do not download again
            model_folder = os.path.join(os.path.dirname(__file__), "model-weights-{}".format(chain))
            os.makedirs(model_folder, exist_ok = True)
            
            if not os.path.isfile(os.path.join(model_folder, "amodel.pt")):
                print("Downloading model ...")
                
                url = "http://opig.stats.ox.ac.uk/website/data/downloads/ablang-{}.tar.gz".format(chain)
                tmp_file = os.path.join(model_folder, "tmp.tar.gz")

                with open(tmp_file,'wb') as f: f.write(requests.get(url).content)
                
                subprocess.run(["tar", "-zxvf", tmp_file, "-C", model_folder], check = True) 
                
                os.remove(tmp_file)

        self.hparams_file = os.path.join(model_folder, 'hparams.json')
        self.model_file = os.path.join(model_folder, 'amodel.pt')
        
        with open(self.hparams_file, 'r', encoding='utf-8') as f:
            self.hparams = argparse.Namespace(**json.load(f))    

        self.AbLang = model.AbLang(self.hparams)
        self.AbLang.to(self.used_device)
        
        if not random_init:
            self.AbLang.load_state_dict(torch.load(self.model_file, map_location=self.used_device))
        
        self.tokenizer = tokenizers.ABtokenizer(os.path.join(model_folder, 'vocab.json'))
        self.AbRep = self.AbLang.AbRep
        
        self.ncpu = ncpu
        self.spread = 11 # Based on get_spread_sequences function
        if chain == 'heavy':
            self.max_position = 128
        else:
            self.max_position = 127
        
        
    def freeze(self):
        self.AbLang.eval()
        
    def unfreeze(self):
        self.AbLang.train()
        
    def __call__(self, sequence, mode='seqcoding', align=False, splitSize=50):
        """
        Mode: sequence, residue, restore or likelihood.
        """
        if not mode in ['rescoding', 'seqcoding', 'restore', 'likelihood']:
            raise SyntaxError("Given mode doesn't exist.")
        
        if isinstance(sequence, str): sequence = [sequence]
        
        
        if align and mode=='restore':
            sequence = self.sequence_aligning(sequence)
            splitSize = ((splitSize//self.spread)+1)*self.spread
        
        aList = []
        for sequence_part in [sequence[x:x+splitSize] for x in range(0, len(sequence), splitSize)]:
            aList.append(getattr(self, mode)(sequence_part, align))
        
        if mode == 'rescoding':
            if align==True:
                return aList
            
            return sum(aList, [])
        
        return np.concatenate(aList)
    
    def seqcoding(self, seqs, align=False):
        """
        Sequence specific representations
        """
        
        tokens = self.tokenizer(seqs, pad=True, device=self.used_device)

        residue_states = self.AbRep(tokens).last_hidden_states
        
        if torch.is_tensor(residue_states): residue_states = residue_states.cpu().detach().numpy()
        
        lens = np.vectorize(len)(seqs)
        
        lens = np.tile(lens.reshape(-1,1,1), (residue_states.shape[2], 1))
        
        seq_codings = np.apply_along_axis(res_to_seq, 2, np.c_[np.swapaxes(residue_states,1,2), lens])
        
        del lens
        del residue_states
        
        return seq_codings
    
    def restore(self, seqs, align=False):
        """
        Restore sequences
        """

        if align:
            nr_seqs = len(seqs)//self.spread

            tokens = self.tokenizer(seqs, pad=True, device=self.used_device)
            predictions = self.AbLang(tokens)[:,:,1:21]
            
            # Reshape
            tokens = tokens.reshape(nr_seqs, self.spread, -1)
            predictions = predictions.reshape(nr_seqs, self.spread, -1, 20)
            seqs = seqs.reshape(nr_seqs, -1)
            
            # Find index of best predictions
            best_seq_idx = torch.argmax(torch.max(predictions, -1).values[:,:,1:2].mean(2), -1)           
            
            # Select best predictions           
            tokens = tokens.gather(1, best_seq_idx.view(-1, 1).unsqueeze(1).repeat(1, 1, tokens.shape[-1])).squeeze(1)
            predictions = predictions[range(predictions.shape[0]), best_seq_idx]
            seqs = np.take_along_axis(seqs, best_seq_idx.view(-1, 1).cpu().numpy(), axis=1)


        else:
            tokens = self.tokenizer(seqs, pad=True, device=self.used_device)
            predictions = self.AbLang(tokens)[:,:,1:21]
            
        predicted_tokens = torch.max(predictions, -1).indices + 1
        restored_tokens = torch.where(tokens==23, predicted_tokens, tokens)

        restored_seqs = self.tokenizer(restored_tokens, encode=False)

        return np.array([res_to_seq(seq, 'reconstruct') for seq in np.c_[restored_seqs, np.vectorize(len)(seqs)]])
    
    def likelihood(self, seqs, align=False):
        """
        Possible Mutations
        """
        
        tokens = self.tokenizer(seqs, pad=True, device=self.used_device)
        
        predictions = self.AbLang(tokens)[:,:,1:21]
        
        if torch.is_tensor(predictions): predictions = predictions.cpu().detach().numpy()

        return predictions
    
    def rescoding(self, seqs, align=False):
        """
        Residue specific representations.
        """
           
        if align:
            
            import pandas as pd
            import anarci
            
            anarci_out = anarci.run_anarci(pd.DataFrame(seqs).reset_index().values.tolist(), ncpu=7, scheme='imgt')
            number_alignment = get_number_alignment(anarci_out)
            
            seqs = np.array([''.join([i[1] for i in onarci[0][0]]).replace('-','') for onarci in anarci_out[1]])
            
            tokens = self.tokenizer(seqs, pad=True, device=self.used_device)
            residue_states = self.AbRep(tokens).last_hidden_states
            
            if torch.is_tensor(residue_states): residue_states = residue_states.cpu().detach().numpy()
            
            residue_output = np.array([create_alignment(res_embed, oanarci, seq, number_alignment) for res_embed, oanarci, seq in zip(residue_states, anarci_out[1], seqs)])
            del residue_states
            del tokens
            
            return output(aligned_embeds=residue_output, number_alignment=number_alignment.apply(lambda x: '{}{}'.format(*x[0]), axis=1).values)
            
        else:
            
            tokens = self.tokenizer(seqs, pad=True, device=self.used_device)
            residue_states = self.AbRep(tokens).last_hidden_states
        
            if torch.is_tensor(residue_states): residue_states = residue_states.cpu().detach().numpy()
            
            residue_output = [res_to_list(state, seq) for state, seq in zip(residue_states, seqs)]

            return residue_output
        
    def sequence_aligning(self, seqs):
        
        import pandas as pd
        import anarci

        anarci_out = anarci.run_anarci(pd.DataFrame(seqs).reset_index().values.tolist(), ncpu=self.ncpu, scheme='imgt') #, allowed_species=['human', 'mouse']
        anarci_data = pd.DataFrame([str(anarci[0][0]) if anarci else 'ANARCI_error' for anarci in anarci_out[1]], columns=['anarci']).astype('<U90')

        seqs = anarci_data.apply(lambda x: get_sequences_from_anarci(x.anarci, 
                                                                     self.max_position, 
                                                                     self.spread), axis=1, result_type='expand').to_numpy().reshape(-1)
        
        return seqs
        
        
            
            

@dataclass
class output():
    """
    Dataclass used to store output.
    """

    aligned_embeds: None
    number_alignment: None
    
    
def res_to_list(state, seq):
    return state[1:1+len(seq)]

def res_to_seq(a, mode='mean'):
    """
    Function for how we go from n_values for each amino acid to n_values for each sequence.
    
    We leave out the start, end and padding tokens.
    """
    if mode=='sum':
        return a[1:(1+int(a[-1]))].sum()
    
    elif mode=='mean':
        return a[1:(1+int(a[-1]))].mean()
    
    elif mode=='reconstruct':
        
        return a[0][1:(1+int(a[-1]))]

def get_number_alignment(oanarci):
    """
    Creates a number alignment from the anarci results.
    """
    
    import pandas as pd
    
    alist = []
    
    for aligned_seq in oanarci[1]:
        alist.append(pd.DataFrame(aligned_seq[0][0])[0])

    unsorted_alignment = pd.concat(alist).drop_duplicates()
    max_alignment = get_max_alignment()
    
    return max_alignment.merge(unsorted_alignment.to_frame(), left_on=0, right_on=0)

def get_max_alignment():
    """
    Create maximum possible alignment for sorting
    """
    
    import pandas as pd

    sortlist = []

    for num in range(1, 128+1):

        if num==112:
            for char in string.ascii_uppercase[::-1]:
                sortlist.append([(num, char)])

            sortlist.append([(num,' ')])

        else:
            sortlist.append([(num,' ')])
            for char in string.ascii_uppercase:
                sortlist.append([(num, char)])
                
    return pd.DataFrame(sortlist)


def create_alignment(res_embeds, oanarci, seq, number_alignment):
    
    import pandas as pd

    datadf = pd.DataFrame(oanarci[0][0])

    sequence_alignment = number_alignment.merge(datadf, how='left', on=0).fillna('-')[1]
        
    idxs = np.where(sequence_alignment.values == '-')[0]

    idxs = [idx-num for num, idx in enumerate(idxs)]

    aligned_embeds = pd.DataFrame(np.insert(res_embeds[1:1+len(seq)], idxs , 0, axis=0))

    return pd.concat([aligned_embeds, sequence_alignment], axis=1).values

def turn_into_numba(anarcis):
    """
    Turns the nested anarci dictionary into a numba item, allowing us to use numba on it.
    """
    
    anarci_list = List.empty_list(unicode_type)
    [anarci_list.append(str(anarci)) for anarci in anarcis]

    return anarci_list

@jit(nopython=True)
def get_spread_sequences(seq, spread, start_position, numbaList):
    """
    Test sequences which are 8 positions shorter (position 10 + max CDR1 gap of 7) up to 2 positions longer (possible insertions).
    """

    for diff in range(start_position-8, start_position+2+1):
        numbaList.append('*'*diff+seq)
    
    return numbaList

def get_sequences_from_anarci(out_anarci, max_position, spread):
    """
    Ensures correct masking on each side of sequence
    """
    
    if out_anarci == 'ANARCI_error':
        return np.array(['ANARCI-ERR']*spread)
    
    end_position = int(re.search(r'\d+', out_anarci[::-1]).group()[::-1])
    # Fixes ANARCI error of poor numbering of the CDR1 region
    start_position = int(re.search(r'\d+,\s\'.\'\),\s\'[^-]+\'\),\s\(\(\d+,\s\'.\'\),\s\'[^-]+\'\),\s\(\(\d+,\s\'.\'\),\s\'[^-]+\'\),\s\(\(\d+,\s\'.\'\),\s\'[^-]+',
                                   out_anarci).group().split(',')[0]) - 1
    
    sequence = "".join(re.findall(r"(?i)[A-Z*]", "".join(re.findall(r'\),\s\'[A-Z*]', out_anarci))))

    sequence_j = ''.join(sequence).replace('-','') + '*'*(max_position-int(end_position))
    
    numba_list = List.empty_list(unicode_type)

    spread_seqs = np.array(get_spread_sequences(sequence_j, spread, start_position, numba_list))

    return spread_seqs





import torch

from .extra_fns import ACT2FN
from .encoderblocks import EncoderBlocks
from .embedding import AbEmbeddings


class AbLang(torch.nn.Module):
    """
    Pretraining model includes Abrep and the head model used for training.
    """
    def __init__(self, hparams):
        super().__init__()
        self.hparams = hparams
        
        self.AbRep = AbRep(self.hparams)       
        self.AbHead = AbHead(self.hparams)
        
    def forward(self, x, attention_mask=None):
        
        representations = self.AbRep(x, attention_mask)
        
        output = self.AbHead(representations.last_hidden_states)
        
        return output
    
    def get_aa_embeddings(self):
        "This function is used to extract the trained aa_embeddings."
        return self.AbRep.AbEmbeddings.aa_embeddings#().weight.detach()

    
class AbRep(torch.nn.Module):
    """
    This is the AbRep model.
    """
    def __init__(self, hparams):
        super().__init__()
        self.hparams = hparams
        
        self.AbEmbeddings = AbEmbeddings(self.hparams)    
        self.EncoderBlocks = EncoderBlocks(self.hparams)
        
        self.init_weights()
        
    def forward(self, src, attention_mask=None, output_attentions=False):
        
        attention_mask = torch.zeros(*src.shape, device=src.device).masked_fill(src == self.hparams.pad_token_id, 1)

        src = self.AbEmbeddings(src)
        
        output = self.EncoderBlocks(src, attention_mask=attention_mask, output_attentions=output_attentions)
        
        return output
    
    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=self.hparams.initializer_range)
        elif isinstance(module, torch.nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            module.bias.data.zero_()
            
    def init_weights(self):
        """
        Initializes and prunes weights if needed.
        """
        # Initialize weights
        self.apply(self._init_weights)
    

class AbHead(torch.nn.Module):
    """
    Head for masked sequence prediction.
    """

    def __init__(self, hparams):
        super().__init__()
        self.hparams = hparams
        self.dense = torch.nn.Linear(self.hparams.hidden_size, self.hparams.hidden_size)
        self.layer_norm = torch.nn.LayerNorm(self.hparams.hidden_size, eps=self.hparams.layer_norm_eps)

        self.decoder = torch.nn.Linear(self.hparams.hidden_size, self.hparams.vocab_size, bias=False)
        self.bias = torch.nn.Parameter(torch.zeros(self.hparams.vocab_size))
        
        self.activation = ACT2FN[self.hparams.hidden_act]
        
        ## self.init_weights() - need to have a function doing this

        self.decoder.bias = self.bias # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`

    def forward(self, features, **kwargs):
        x = self.dense(features)

        x = self.activation(x)
        x = self.layer_norm(x)

        # project back to size of vocabulary with bias
        x = self.decoder(x)

        return x



import torch


class AbEmbeddings(torch.nn.Module):
    """
    Residue embedding and Positional embedding
    """
    
    def __init__(self, hparams):
        super().__init__()
        self.pad_token_id = hparams.pad_token_id
        
        self.AAEmbeddings = torch.nn.Embedding(hparams.vocab_size, hparams.hidden_size, padding_idx=self.pad_token_id)
        self.PositionEmbeddings = torch.nn.Embedding(hparams.max_position_embeddings, hparams.hidden_size, padding_idx=0) # here padding_idx is always 0
        
        self.LayerNorm = torch.nn.LayerNorm(hparams.hidden_size, eps=hparams.layer_norm_eps)
        self.Dropout = torch.nn.Dropout(hparams.hidden_dropout_prob)

    def forward(self, src):
        
        inputs_embeds = self.AAEmbeddings(src)
        
        position_ids = self.create_position_ids_from_input_ids(src, self.pad_token_id)   
        position_embeddings = self.PositionEmbeddings(position_ids)

        embeddings = inputs_embeds + position_embeddings

        return self.Dropout(self.LayerNorm(embeddings))
        
    def create_position_ids_from_input_ids(self, input_ids, padding_idx):
        """
        Replace non-padding symbols with their position numbers. Padding idx will get position 0, which will be ignored later on.
        """
        mask = input_ids.ne(padding_idx).int()
        
        return torch.cumsum(mask, dim=1).long() * mask


import json
import torch

class ABtokenizer():
    """
    Tokenizer for proteins. Both aa to token and token to aa.
    """
    
    def __init__(self, vocab_dir):
        self.set_vocabs(vocab_dir)
        self.pad_token = self.vocab_to_token['-']
        
    def __call__(self, sequenceList, encode=True, pad=False, device='cpu'):
        #assert isinstance(sequenceList, list)
        
        if encode: 
            data = [self.encode(seq, device=device) for seq in sequenceList]
            if pad: return torch.nn.utils.rnn.pad_sequence(data, batch_first=True, padding_value=self.pad_token)
            else: return data
        
        else: return [self.decode(token) for token in sequenceList]
    
    def set_vocabs(self, vocab_dir):
        with open(vocab_dir, encoding="utf-8") as vocab_handle:
            self.vocab_to_token=json.load(vocab_handle)
            
        self.vocab_to_aa = {v: k for k, v in self.vocab_to_token.items()}
     
    def encode(self, sequence, device='cpu'):
        
        encoded = [self.vocab_to_token["<"]]+[self.vocab_to_token[resn] for resn in sequence]+[self.vocab_to_token[">"]]
        
        return torch.tensor(encoded, dtype=torch.long, device=device)
        # Start and Stop token should probably not be added here, but instead earlier
    
    def decode(self, seqtokens):
        
        if torch.is_tensor(seqtokens): seqtokens = seqtokens.cpu().numpy()

        return ''.join([self.vocab_to_aa[token] for token in seqtokens])
    

    



import math
from typing import List, Optional, Tuple
from dataclasses import dataclass

import torch
import torch.nn as nn
from fairseq.modules.multihead_attention import MultiheadAttention

from .extra_fns import ACT2FN


@dataclass
class AbRepOutput():
    """
    Dataclass used to store AbRep output.
    """

    last_hidden_states: torch.FloatTensor
    all_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[torch.FloatTensor]] = None

        
class EncoderBlocks(torch.nn.Module):
    """
    Wrapper for multiple EncoderBlocks (or a single).
    """
    def __init__(self, hparams):
        super().__init__()
        self.hparams = hparams
        self.Layers = nn.ModuleList([EncoderBlock(hparams) for _ in range(hparams.num_hidden_layers)])

    def forward(self, hidden_states, attention_mask=None, output_attentions=False, output_hidden_states=False):
        
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        
        for num_block, a_EncoderBlock in enumerate(self.Layers):
            
            hidden_states, attentions = a_EncoderBlock(hidden_states, attention_mask, output_attentions)
            #print(attentions)
            
            if output_hidden_states: 
                all_hidden_states = all_hidden_states + (hidden_states,) # Takes out each hidden states after each EncoderBlock
            
            if output_attentions: 
                all_self_attentions = all_self_attentions + (attentions,) # Takes out attention layers for analysis
                  
        return AbRepOutput(last_hidden_states=hidden_states, all_hidden_states=all_hidden_states, attentions=all_self_attentions)
    
    
class EncoderBlock(torch.nn.Module):
    """
    Single EncoderBlock.
    
    An EncoderBlock consists of a MultiHeadAttention and a IntermediateLayer.
    """
    def __init__(self, hparams):
        super().__init__()
        
        self.MultiHeadAttention = ThirdMultiHeadAttention(hparams)
        self.MHADropout = nn.Dropout(hparams.hidden_dropout_prob)
        self.MHALayerNorm = nn.LayerNorm(hparams.hidden_size, eps=hparams.layer_norm_eps)
        
        self.IntermediateLayer = IntermediateLayer(hparams)
        
    def forward(self, hidden_states, attention_mask=None, output_attentions=False):

        MHAoutput, attentions = self.MultiHeadAttention(hidden_states, attention_mask, output_attentions=output_attentions)
         
        output = self.MHADropout(MHAoutput)
        output = self.MHALayerNorm(output + hidden_states) # HIDDEN_STATES ARE ADDED FOR RESIDUAL BLOCK EFFECT
        
        output = self.IntermediateLayer(output) # INTERMEDIATELAYER HAS RESIDUAL BLOCK EFFECT INTERNALLY

        #outputs = (layer_output,) + self_attention_outputs[1:]  # if output_attentions=False then 1: is empty
        
        return output, attentions
    
    
class ThirdMultiHeadAttention(torch.nn.Module):
    """
    New MultiHeadAttention which can return the weights of the individual heads.
    """
    
    def __init__(self, hparams):
        super().__init__()
                
        self.Attention = MultiheadAttention(hparams.hidden_size, hparams.num_attention_heads, dropout=hparams.attention_probs_dropout_prob, self_attention=True)
        
    def forward(self, hidden_states, attention_mask=None, output_attentions=False):
                
        hidden_states = torch.transpose(hidden_states, 0, 1)
        
        # static_kv is only True because there is currently a bug which doesn't return the head weights unaveraged unless its true
        attn_output, attn_weights = self.Attention(hidden_states, hidden_states, hidden_states, key_padding_mask=attention_mask, static_kv=True, 
                                                   need_weights=output_attentions, need_head_weights=output_attentions)

        return torch.transpose(attn_output, 0, 1), attn_weights


class OldMultiHeadAttention(torch.nn.Module):
    """
    MultiHeadAttention contains a Scaled Dot Product Attention and a Linear Layer.    
    """
    def __init__(self, config):
        super().__init__()
        self.Attention = torch.nn.MultiheadAttention(config.hidden_size, config.num_attention_heads, config.attention_probs_dropout_prob)

    def forward(self, hidden_states, attention_mask=None, output_attentions=False):  
        
        hidden_states = torch.transpose(hidden_states, 0, 1)
        output, attentions = self.Attention(hidden_states, hidden_states, hidden_states, key_padding_mask=attention_mask, need_weights=output_attentions) 
        
        attention_output = torch.transpose(output, 0, 1)

        return attention_output, attentions
    

class IntermediateLayer(nn.Module):
    """
    Contains an expanding layer, while also functioning as a residual block ending with a drop-norm layer
    """
    def __init__(self, config):
        super().__init__()
        self.expand_dense = nn.Linear(config.hidden_size, config.intermediate_size)
        self.intermediate_act_fn = ACT2FN[config.hidden_act]
        
        self.dense_dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(self, hidden_states):
        output = self.expand_dense(hidden_states)
        output = self.intermediate_act_fn(output)
               
        output = self.dense_dense(output)
        output = self.dropout(output)
        output = self.LayerNorm(output + hidden_states)
                
        return output 



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

    
setup(
    name='ablang',
    version='0.2.2',
    description='AbLang: A language model for antibodies.',
    license='BSD 3-clause license',
    maintainer='Tobias Hegelund Olsen',
    long_description=long_description,
    long_description_content_type='text/markdown',
    maintainer_email='tobias.olsen@stats.ox.ac.uk',
    include_package_data=True,
    packages=find_packages(include=('ablang', 'ablang.*')),
    install_requires=[
        'numpy',
        'numba',
        'requests',
        'fairseq',
        'torch>=1.6',
    ],
)